#include "MetaGraph.hh"

double MetaGraph::estimateCoverageSubgraph(int cluster)
{

	IDnum index;
	long count=0;
        // Write nodes into array and compute total assembly length
        for (index = 1; index <= nodeCount(graph); index++) {
                if (subgraphMask[index+nodeCount(graph)]==cluster) {
                        count++;        
                }
        }

	//cout<<"number of element in cluster "<<count<<endl;
	Node ** nodeArray = callocOrExit(count, Node*); 
	Coordinate halfTotalLength = 0;
	Coordinate sumLength = 0;
	Node *node;
	
	if (nodeCount(graph) == 0) {
		cout<<"EMPTY GRAPH"<<endl;
		return 0;
	}

		
	count=0;
	// Write nodes into array and compute total assembly length
	for (index = 1; index <= nodeCount(graph); index++) {
		if (subgraphMask[index+nodeCount(graph)]==cluster) {
			node = getNodeInGraph(graph, index);
			nodeArray[count] = node;
			count++;
			if (node == NULL)
				continue;
			halfTotalLength += getNodeLength(node);
		}
	}
	halfTotalLength /= 2;
	
	//cout<<"done finish copy into array "<<count<<endl;

	// Sort nodes
	qsort(nodeArray, count, sizeof(Node *), compareNodeCovs);

	//cout<<"done finish qsort"<<endl;
	
	count=0;
	// Compute the length weighted median node coverage
	for (index = 1; index <= nodeCount(graph); index++) {
		if (subgraphMask[index+nodeCount(graph)]==cluster) {
			node = nodeArray[count];
			count++;
			sumLength += getNodeLength(node);
			if (sumLength >= halfTotalLength) {
			  	free(nodeArray);
				return getTotalCoverage(node) / (double) getNodeLength(node);
			}
		}
	}
	cout<<"finish return"<<endl;

	// In case something went wrong...
	free(nodeArray);
	return -1;	
}

void MetaGraph::setUniqueTrue() {
  IDnum index;
  Node *node;
  for(index = 0; index < nodeCount(graph); index++) {
    node = getNodeInGraph(graph, index + 1);
    if(node == NULL) {
      continue;
    }
      setUniqueness(node, true);   
  }
}

void MetaGraph::estimateBestCoverage(){
	bestCoverage=callocOrExit( cluster , double );
	for (int i=0;i<cluster;i++){		
		bestCoverage[i]=estimateCoverageSubgraph(i+1);
	}
	// Sort cluster based on coverage
	qsort(bestCoverage, cluster, sizeof(double *), compareSubgraphCov);
	cout<<"highest cov "<<bestCoverage[0]<<endl;
        cout<<"highest cov "<<bestCoverage[1]<<endl;
        cout<<"highest cov "<<bestCoverage[2]<<endl;
        cout<<"highest cov "<<bestCoverage[3]<<endl;
        cout<<"highest cov "<<bestCoverage[4]<<endl;
}

int MetaGraph::compareSubgraphCov(const void * a,const void * b){
	if (*(double*)a==*(double*)b) return 0;
	else if (*(double*)a > *(double*)b) return -1;
	else return 1;	
}

double MetaGraph::coverageCutOff(){
	return bestCoverage[cluster-1]/2;	
}


IDnum MetaGraph::getNumUniqueNodes() const {
  IDnum count = 0;
  for( IDnum i=1 ; i<=nodeCount(graph) ; i++ ){
    Node* node = getNodeInGraph( graph, i );
    if (getUniqueness( node )){
      count++;
    }
  }
  return count;
}



void MetaGraph::identifyUniqueNodesSubgraphOnce( double Coverage ){
  for( IDnum i=0; i<nodeCount(graph) ; ++i ){
    Node* node = getNodeInGraph( graph, i+1 );
    if( node != NULL ){
      if( subgraphMask[i+1+nodeCount(graph)] != META_GRAPH_MASK_UNIQUE ){
	setUniqueness( node, isUniqueSubgraph(node, Coverage) );
	if (getUniqueness(node)){
		subgraphMask[i+1+nodeCount(graph)] = META_GRAPH_MASK_UNIQUE;
			FastaSeq* seqT2 = FastaSeq::instantiate( VUtils::getTitle(node));
			writerUniqueNode->send( seqT2 );
			delete seqT2;
			uniqueNode++;
	 	}
      } else {
	setUniqueness(node, false);
      }
    }
  }
}


void MetaGraph::scaffoldingWithMultiPeakModeUniqueOnce( boolean* flagMatePair, bool flagScaffolding){
	//pake
  cout << "[MetaGraph] " << " === Create read paring array ===" << endl;
  createReadPairingArray(sequences);

  cout << "[MetaGraph] " << " === Detach dubious reads ===" << endl;
  detachDubiousReads( sequences, dubious );

  cout << "[MetaGraph] " << " === Activate gap markers ===" << endl;
  activateGapMarkers(graph);

  cout << "[MetaGraph] " << " === Split into subgraphs ===" << endl;

  bool flagLongRead = checkLongReadExistence();
  subgraphMask = callocOrExit( getSubgraphMaskSize() , int );
  eliminateNullNodes();

	 
        writerUniqueNode = new FastaWriter( "[UniqueNodeA7]" );
	writerUniqueNode->open( subgraphPrefix + "__UniqueNodeA7New");
	uniqueNode=0;

	int pembagi=(cluster-10)/10;

for( int i = cluster; i>=10 ; i--) {
        if ((i%pembagi==0)|(i==cluster)){
        double coverage = bestCoverage[i];
        if (coverage <= 0)continue;
        resetNodeFlags();
        cout << "[MetaGraph] " << " === Subgraph contiging === " <<i<<" "<<bestCoverage[i]<<endl;
        identifyUniqueNodesSubgraphOnce( coverage );
	saveSubgraph( i, coverage );

        if( flagLongRead ){
        cout << "[MetaGraph] " << " === Subgraph Rock Band ===" << endl;
        readCoherentSubgraph( coverage );
        }

        cout << "[MetaGraph] " << " === Subgraph Scaffolding ===" << endl;
        for ( int pebbleRounds=pairedCategories(sequences)+1 ; pebbleRounds>0 ; --pebbleRounds){
                exploitShortReadPairs( graph, sequences, dubious, flagMatePair, flagScaffolding );
        }
        cout << "[MetaGraph] " << " === Done: Subgraph contiging & scaffolding === " << endl << endl;

        resetUniqueness();
        eliminateNullNodes();
      }
}




for( int i = 9; i>=0 ; i--) {
	double coverage = bestCoverage[i];
	if (coverage <= 0)continue;
	resetNodeFlags();
    	cout << "[MetaGraph] " << " === Subgraph contiging === " <<i<<" "<<bestCoverage[i]<<endl;
    	identifyUniqueNodesGraph(coverage);
    	saveSubgraph( i, coverage );
	
	if( flagLongRead ){
      	cout << "[MetaGraph] " << " === Subgraph Rock Band ===" << endl;
      	readCoherentSubgraph( coverage );
    	}

	cout << "[MetaGraph] " << " === Subgraph Scaffolding ===" << endl;
    	for ( int pebbleRounds=pairedCategories(sequences)+1 ; pebbleRounds>0 ; --pebbleRounds){
      		exploitShortReadPairs( graph, sequences, dubious, flagMatePair, flagScaffolding );
    	}
    	cout << "[MetaGraph] " << " === Done: Subgraph contiging & scaffolding === " << endl << endl;

        resetUniqueness();
	eliminateNullNodes();

}
	writerUniqueNode->close();
	delete writerUniqueNode;
	cout<<"Unique Node= "<<uniqueNode<<endl;
}   


void MetaGraph::identifyUniqueNodesGraph( double primaryCoverage ){
  for( IDnum i=0; i<nodeCount(graph) ; ++i ){
    Node* node = getNodeInGraph( graph, i+1 );
    if( node != NULL ){
	setUniqueness( node, isUniqueSubgraph(node, primaryCoverage) );
      } 
    }
  }


void MetaGraph::identifyUniqueNodesSubgraph( double primaryCoverage, int cluster ){
  for( IDnum i=0; i<nodeCount(graph) ; ++i ){
    Node* node = getNodeInGraph( graph, i+1 );
    if( node != NULL ){
      if( subgraphMask[i+1+nodeCount(graph)] == cluster ){
	setUniqueness( node, isUniqueSubgraph(node, primaryCoverage) );
      } else {
	setUniqueness(node, false);
      }
    }
  }
}

int MetaGraph::extractChimeraNodeStruct(boolean* flagMatePair){

        FastaWriter* writer = new FastaWriter( "[ChimeraNodeStruct]" );
        writer->open( subgraphPrefix + "__ChimeraNodeCandidates");

        FastaWriter* writerTitle = new FastaWriter( "[TitleChimeraNodeStruct]" );
        writerTitle->open( subgraphPrefix + "__TitleChimeraNodeCandidates");

        FastaWriter* writerAlign = new FastaWriter( "[ChimeraNodeStructAlign]" );
        writerAlign->open( subgraphPrefix + "__ChimeraNodeCandidatesAlign");

        
	cout << "[MetaGraph] " << " === Create read paring array ===" << endl;
	createReadPairingArray(sequences);

	cout << "[MetaGraph] " << " === Detach dubious reads ===" << endl;
	detachDubiousReads( sequences, dubious );

	cout << "[MetaGraph] " << " === Activate gap markers ===" << endl;
	activateGapMarkers(graph);

        int num =0;

	

	 if( splitJudge->getFlagUseConnections() ){
     		 setUniqueTrue();
		 _computeConnections( flagMatePair );
	}

	//cout <<"setelah compute connection"<<endl;

		for ( IDnum i=0; i<nodeCount(graph) ; ++i ){
			Node* node = getNodeInGraph( graph, i+1 );
			if( getNodeID(node) == NULL_IDX ){
				continue;
			}

			if( getNodeLength(node) <2 ){
                                continue;
                        }

		if( splitJudge->isLocalRepeatStructure(node) ){
                vector<Node*> inNodes  = VUtils::getInNodes( node );
                vector<Node*> outNodes = VUtils::getOutNodes( node );
                Node* primaryInNode    = VUtils::getMaxDensityNode( inNodes );
		if (getNodeLength(primaryInNode)<11) continue;
                Node* secondaryInNode  = VUtils::getMinDensityNode( inNodes );
		if (getNodeLength(secondaryInNode)<11) continue;
                Node* primaryOutNode   = VUtils::getMaxDensityNode( outNodes );
		if (getNodeLength(primaryOutNode)<11) continue;
                Node* secondaryOutNode = VUtils::getMinDensityNode( outNodes );
		if (getNodeLength(secondaryOutNode)<11) continue;

		

	//	cout<<"before calcnumConnection"<<endl;
		string title=VUtils::getTitle(node)+"___"
	+Utils::itoa(splitJudge->calcNumConnections(primaryInNode, primaryOutNode)+splitJudge->calcNumConnections(getTwinNode(primaryInNode), primaryOutNode))+"___"
	+Utils::ltoa(splitJudge->calcNumConnections(secondaryInNode, secondaryOutNode)+splitJudge->calcNumConnections(getTwinNode(secondaryInNode), secondaryOutNode))+"___"
	+Utils::ltoa(splitJudge->calcNumConnections(primaryInNode, secondaryOutNode)+splitJudge->calcNumConnections(getTwinNode(primaryInNode), secondaryOutNode))+"___"
	+Utils::ltoa(splitJudge->calcNumConnections(secondaryInNode, primaryOutNode)+splitJudge->calcNumConnections(getTwinNode(secondaryInNode), primaryOutNode));


	//	cout<<"after calcnumconnection"<<endl;
                FastaSeq* seq = FastaSeq::instantiate( title, VUtils::getSequence(node) );
                writer->send( seq );
                delete seq;
				
		FastaSeq* seqT = FastaSeq::instantiate(title);
                writerTitle->send( seqT );
                delete seqT;

                FastaSeq* seq1 = FastaSeq::instantiate( VUtils::getTitle(primaryInNode), VUtils::getSequence(primaryInNode) );
                writer->send( seq1 );
		writerAlign->send( seq1 );
                delete seq1;

		FastaSeq* seqT1 = FastaSeq::instantiate( VUtils::getTitle(primaryInNode));                                                   
		writerTitle->send( seqT1 );
                delete seqT1;                                           
              
                FastaSeq* seq2 = FastaSeq::instantiate( VUtils::getTitle(secondaryInNode), VUtils::getSequence(secondaryInNode));
                writer->send( seq2 );     
		writerAlign->send( seq2 );                                                                                    
                delete seq2;                                                                                                  
                                                                                                                              
                                                                                                                          
                FastaSeq* seqT2 = FastaSeq::instantiate( VUtils::getTitle(secondaryInNode));                                  
                writerTitle->send( seqT2 );                                                                                   
                delete seqT2;                                                                                                 
                                                                                                                        
                                                                                                                              
                FastaSeq* seq3 = FastaSeq::instantiate( VUtils::getTitle(primaryOutNode), VUtils::getSequence(primaryOutNode));
                writer->send( seq3 );                                                                                         
		writerAlign->send( seq3 );
                delete seq3;                                                                                                  
                                                                                                                              
                FastaSeq* seqT3 = FastaSeq::instantiate( VUtils::getTitle(primaryOutNode));                                  
                writerTitle->send( seqT3 );                                                                                   
                delete seqT3;                                                                                                 
                                                                                                                            
                FastaSeq* seq4 = FastaSeq::instantiate( VUtils::getTitle(secondaryOutNode), VUtils::getSequence(secondaryOutNode));
                writer->send( seq4 );  
		writerAlign->send( seq4 );                                                                                       
                delete seq4;                                                                                                  
                                                                                                                              
                FastaSeq* seqT4 = FastaSeq::instantiate( VUtils::getTitle(secondaryOutNode));                                
                writerTitle->send( seqT4 );                                                                                   
                delete seqT4;            
        num++;                                                                                                              
        }                                                                                                                   
                                                                                                                            
  }                                                                                                                         
  writer->close();                                                                                                          
  delete writer;                                                                                                            
  writerTitle->close();                                                                                                          
  delete writerTitle;                  
  writerAlign->close();
  delete writerAlign;
                                                                                          
  return num;                                                                                                               
  }                    

void MetaGraph::extractSubgraph(int numbNode){
	subgraphMask = callocOrExit( getSubgraphMaskSize() , int );
	eliminateNullNodes();
	
	bool flagVisitedAllNodes = false;
	int countSubgraph=0;
	cluster=0;
	counter=0;
	while ( IDnum nodeID = getUnvisitedNodeID() ){
			counter=0;
			extractNextSubgraph( nodeID );
			if (counter>numbNode) {
				cluster++;
				changeSubgraphMask( SUB_GRAPH_MASK_NOW_VISITING, cluster );
				counter /=2;							
				//cout<<counter<<" ";
			}
			else 
				changeSubgraphMask( SUB_GRAPH_MASK_NOW_VISITING, SUB_GRAPH_MASK_TRASH);			
			countSubgraph++;
	}
	cout<<endl;
	cout<<"counter > 2= "<<cluster<<endl;
	cout<<"Number of subgraph= "<<countSubgraph<<endl;
	counter=0;
	cout<<"Number of node= "<<nodeCount(graph)<<endl;
}
  
int MetaGraph::splitRepeats(string chimeraNodes[],string classificationChimera[]){
	long numInterRepeats = 0;
	vector<Node*> inNodes, outNodes;
	resetNodeFlags();

	long NodeID=0;
	for (int chimID=0;chimID<75000;++chimID)

	{
		//retrive NodeID from chimera array
		if (chimeraNodes[chimID]==""){ continue;}
		if (classificationChimera[chimID] == "3"){
			continue;
		}

	
		if (classificationChimera[chimID] == "4"){
                        continue;
                }

		char * cstr = new char [chimeraNodes[chimID].length()+1];
		strcpy (cstr, chimeraNodes[chimID].c_str());
		char * p = strtok (cstr,"___");
		NodeID = atol(p);
		delete[] cstr;
		p=NULL;
		delete p;

		
        	//retrieve Node from NodeID
		Node* node = getNodeInGraph(graph,NodeID);
		//cout<<"check retrieve node from nodeID "<<endl;	
	
		if( getNodeID(node) == NULL_IDX ){
			continue;
		}

		if( node == NULL){
                        continue;
                }


		//cout<<"before localrepeatstructure "<<getNodeID(node)<<endl;

       		if(!(splitJudge->isLocalRepeatStructure(node))){
			continue;
       		}

		//cout<<"after localRepeatStructure "<<endl;
		/*if( subgraphMask[getNodeID(node)+nodeCount(graph)] != META_GRAPH_MASK_NOW_VISITING ){
			continue;
		}*/

		//++numInterRepeats;
        	//retrieve primaryNode + secondary
        	vector<Node*> inNodes  = VUtils::getInNodes( node );
        	vector<Node*> outNodes = VUtils::getOutNodes( node );

		//cout<<"cek inNodes and outNodes "<<endl;
		Node* primaryInNode    = VUtils::getMaxDensityNode( inNodes );                                          
		Node* secondaryInNode  = VUtils::getMinDensityNode( inNodes );
		Node* primaryOutNode   = VUtils::getMaxDensityNode( outNodes );
		Node* secondaryOutNode = VUtils::getMinDensityNode( outNodes );

		//cout<<getNodeID(node)<<" "<<getNodeID(primaryInNode)<<" "<<getNodeID(secondaryInNode)<<" "<<getNodeID(primaryOutNode)<<" "<<getNodeID(secondaryOutNode)<<endl;

		if (getNodeID(primaryInNode)==getNodeID(secondaryInNode)) continue;
		if (getNodeID(primaryInNode)==getNodeID(secondaryOutNode)) continue;
		if (getNodeID(primaryInNode)==getNodeID(primaryOutNode)) continue;
		if (getNodeID(primaryInNode)==-(getNodeID(primaryOutNode))) continue;

		if (getNodeID(secondaryInNode)==getNodeID(secondaryOutNode)) continue;
		if (getNodeID(secondaryInNode)==-(getNodeID(secondaryOutNode))) continue;

		if (getNodeID(secondaryInNode)==getNodeID(primaryOutNode)) continue;

		if (getNodeID(primaryOutNode)==getNodeID(secondaryOutNode)) continue;

		setNodeStatus(node, true); setUniqueness(node, false);
        	
		//determine type of split based on classfication
		if (classificationChimera[chimID] == "1") {
			setNodeStatus(primaryInNode, true);
			setUniqueness(primaryInNode, true);
			setNodeStatus(primaryOutNode, true);
			setUniqueness(primaryOutNode, true);
            		pushNeighboursInterRepeat( primaryInNode, node, primaryOutNode );

            		setNodeStatus(secondaryInNode, true);
            		setUniqueness(secondaryInNode, true);
            		setNodeStatus(secondaryOutNode, true);
            		setUniqueness(secondaryOutNode, true);
            		pushNeighboursInterRepeat( secondaryInNode, node, secondaryOutNode );
			destroyNode( node, graph );
            		++numInterRepeats;
        	}
        	else if (classificationChimera[chimID] == "2"){
			setNodeStatus(primaryInNode, true);
            		setUniqueness(primaryInNode, true);
            		setNodeStatus(secondaryOutNode, true);
            		setUniqueness(secondaryOutNode, true);
			pushNeighboursInterRepeat( primaryInNode, node, secondaryOutNode );

            		setNodeStatus(secondaryInNode, true);
            		setUniqueness(secondaryInNode, true);
            		setNodeStatus(primaryOutNode, true);
            		setUniqueness(primaryOutNode, true);
            		pushNeighboursInterRepeat( secondaryInNode, node, primaryOutNode );
			destroyNode( node, graph );
            		++numInterRepeats;
            	}	
        }
  resetNodeFlags();
  return numInterRepeats;
}


void MetaGraph::scaffolding( boolean* flagMatePair, bool flagScaffolding, double maxChimeraRate, bool flagDiscardChimera ){
  if( splitJudge->getNumCoveragePeaks() == 1 ){
    cout << "[MetaGraph] " << " === Scaffolding with single peak mode ===" << endl;
    return scaffoldingWithSinglePeakMode( splitJudge->getExpectedCoverage(0), flagMatePair, flagScaffolding );
  }
  cout << "[MetaGraph] " << " === Scaffolding with multi-peak mode ===" << endl;
  scaffoldingWithMultiPeakMode( flagMatePair, flagScaffolding, maxChimeraRate, flagDiscardChimera );
				
}

void MetaGraph::scaffoldingWithMultiPeakMode( boolean* flagMatePair, bool flagScaffolding, double maxChimeraRate, bool flagDiscardChimera ){
  cout << "[MetaGraph] " << " === Create read paring array ===" << endl;
  createReadPairingArray(sequences);

  cout << "[MetaGraph] " << " === Detach dubious reads ===" << endl;
  detachDubiousReads( sequences, dubious );

  cout << "[MetaGraph] " << " === Activate gap markers ===" << endl;
  cout << "[MetaGraph] " << " === Activate gap markers ===" << endl;
  activateGapMarkers(graph);

  cout << "[MetaGraph] " << " === Split into subgraphs ===" << endl;
  bool flagLongRead = checkLongReadExistence();
  subgraphMask = callocOrExit( getSubgraphMaskSize() , int );
  eliminateNullNodes();

  size_t roundID = 0;
  size_t primaryPeakID = 0;
  bool flagVisitedAllNodes = false;
  while( !flagVisitedAllNodes ){
    double primaryCoverage   = splitJudge->getExpectedCoverage(primaryPeakID);
    double secondaryCoverage = splitJudge->getExpectedCoverage(primaryPeakID+1);
    cout << "[MetaGraph] " << (roundID+1) << "-th round" << endl
	 << "[MetaGraph] " << "Primary coverage   = " << primaryCoverage << endl
	 << "[MetaGraph] " << "Secondary coverage = " << secondaryCoverage << endl;
    if( secondaryCoverage > 0 ){
      ++primaryPeakID;
    }
    ++roundID;

    if( splitJudge->getFlagUseConnections() ){
      setSubgraphUniqueness();
      _computeConnections( flagMatePair );
    }

    while( IDnum nodeID = getUnvisitedNodeID() ){
      extractNextSubgraph( nodeID );
      if( isChimeraSubgraph( primaryCoverage, secondaryCoverage, maxChimeraRate ) ){
	splitRepeats();
	eliminateNullNodes();
	if( flagDiscardChimera ){
	  changeSubgraphMask( META_GRAPH_MASK_NOW_VISITING, META_GRAPH_MASK_DELETED );
	} else {
	  forceSeparateChimericSubgraph( primaryCoverage, secondaryCoverage );
	}
      } else if( isPrimarySubgraph( primaryCoverage, secondaryCoverage ) ){
	changeSubgraphMask( META_GRAPH_MASK_NOW_VISITING, META_GRAPH_MASK_PRIMARY );
      }
    }

    resetNodeFlags();
    cout << "[MetaGraph] " << " === Subgraph contiging ===" << endl;
    identifyUniqueNodesSubgraph( primaryCoverage ); 
    saveSubgraph( roundID, primaryCoverage );
    
    if( flagLongRead ){
      cout << "[MetaGraph] " << " === Subgraph Rock Band ===" << endl;
      readCoherentSubgraph( primaryCoverage );
    }
    
    cout << "[MetaGraph] " << " === Subgraph Scaffolding ===" << endl;
    for ( int pebbleRounds=pairedCategories(sequences)+1 ; pebbleRounds>0 ; --pebbleRounds){
      exploitShortReadPairs( graph, sequences, dubious, flagMatePair, flagScaffolding );
    }
    cout << "[MetaGraph] " << " === Done: Subgraph contiging & scaffolding === " << endl << endl;
    
    eliminateNullNodes();
    resetUniqueness();
    changeSubgraphMask( META_GRAPH_MASK_PRIMARY, META_GRAPH_MASK_DELETED );
    changeSubgraphMask( META_GRAPH_MASK_SECONDARY, META_GRAPH_MASK_UNVISITED );
    if( getUnvisitedNodeID() == 0 ){
      flagVisitedAllNodes = true;
    }
  }
  free(subgraphMask);	
}



void MetaGraph::eliminateNullNodes(){
  for( IDnum i = 0; i<getSubgraphMaskSize() ; ++i ) {
    Node* node = getNodeInGraph( graph, i-nodeCount(graph) );
    if( node == NULL || getNodeID(node)==0 ){
      subgraphMask[i] = META_GRAPH_MASK_DELETED;
    }
  }
}

void MetaGraph::setSubgraphUniqueness() {
  IDnum index;
  Node *node;
  for(index = 0; index < nodeCount(graph); index++) {
    node = getNodeInGraph(graph, index + 1);
    if(node == NULL) {
      continue;
    }
    if(subgraphMask[index + 1 + nodeCount(graph)] == META_GRAPH_MASK_UNVISITED) {
      setUniqueness(node, true);
    } else {
      setUniqueness(node, false);
    }
  }
}

void MetaGraph::_computeConnections( boolean* flagMatePair ) {
  computeConnections( graph, sequences, dubious, flagMatePair );
}

void MetaGraph::extractNextSubgraph( IDnum nodeID ){
  dfs( nodeID );
  dfs( -1 * nodeID );
}

IDnum MetaGraph::getUnvisitedNodeID() const {
  long maskSize = getSubgraphMaskSize();
  for( long i=0 ; i<maskSize ; ++i ){
    if( i == (maskSize-1)/2 ){
      continue;
    }
    if( subgraphMask[i] == SUB_GRAPH_MASK_UNVISITED ){
      return i - (maskSize-1) / 2;
    }
  }
  return NULL_IDX;
}

void MetaGraph::dfs( IDnum currentIndex ){
  if( subgraphMask[currentIndex + nodeCount(graph)] == SUB_GRAPH_MASK_UNVISITED ){
    subgraphMask[currentIndex + nodeCount(graph)] = SUB_GRAPH_MASK_NOW_VISITING; 
  //  cout<<"!";
	counter++;
    Arc *activeArc = NULL;
    for( activeArc=getArc(getNodeInGraph(graph, currentIndex)) ; activeArc!=NULL ; activeArc=getNextArc(activeArc) ){
      long nextIndex = getNodeID(getDestination(activeArc));
      if( subgraphMask[nextIndex+nodeCount(graph)] == SUB_GRAPH_MASK_UNVISITED ){
	dfs( nextIndex );
	dfs( -1 * nextIndex );
      }
    }
  }
}

bool MetaGraph::isChimeraSubgraph( double primaryCoverage, double secondaryCoverage, double maxChimeraRate ) const {
  double totalPrimaryLength = 0.0;
  double totalSecondaryLength = 0.0;
  for ( long i=1 ; i<=nodeCount(graph) ; ++i ){
    if( subgraphMask[i + nodeCount(graph)] == META_GRAPH_MASK_NOW_VISITING ){
      Node* node = getNodeInGraph( graph, i );
      if( node == NULL )
	continue;
      double curCoverage = VUtils::getNodeDensity( node );
      if ( fabs(curCoverage-primaryCoverage) <= fabs(curCoverage-secondaryCoverage) ){
	totalPrimaryLength += getNodeLength( node );
      } else {
	totalSecondaryLength += getNodeLength( node );
      }
    }
  }
  double primaryFraction = totalPrimaryLength / (totalPrimaryLength + totalSecondaryLength);
  double secondaryFraction = totalSecondaryLength / (totalPrimaryLength + totalSecondaryLength);
  if( primaryFraction>maxChimeraRate and secondaryFraction>maxChimeraRate ){
    return true;
  }
  return false;
}

bool MetaGraph::isPrimarySubgraph( double primaryCoverage, double secondaryCoverage ) const {
  double totalPrimaryLength = 0.0;
  double totalSecondaryLength = 0.0;
  for ( long i=1 ; i<=nodeCount(graph) ; ++i ){
    if( subgraphMask[i + nodeCount(graph)] == META_GRAPH_MASK_NOW_VISITING ){
      Node* node = getNodeInGraph( graph, i );
      if( node == NULL )
	continue;
      double curCoverage = VUtils::getNodeDensity( node );
      if( fabs(curCoverage - primaryCoverage) <= fabs(curCoverage-secondaryCoverage) ){
	totalPrimaryLength += getNodeLength( node );
      } else {
	totalSecondaryLength += getNodeLength( node );
      }
    }
  }
  double primaryFraction = totalPrimaryLength / (totalPrimaryLength + totalSecondaryLength);
  double secondaryFraction = totalSecondaryLength / (totalPrimaryLength + totalSecondaryLength);
  if( primaryFraction > secondaryFraction ){
    return true;
  }
  return false;
}

int MetaGraph::splitRepeats(){
  int numInterRepeats = 0;
  vector<Node*> inNodes, outNodes;
  resetNodeFlags();
  for ( IDnum i=0; i<nodeCount(graph) ; ++i ){
    Node* node = getNodeInGraph( graph, i+1 );
    if( getNodeID(node) == NULL_IDX ){
      continue;
    }
    if( subgraphMask[getNodeID(node)+nodeCount(graph)] != META_GRAPH_MASK_NOW_VISITING ){
      continue;
    }

    if( splitJudge->isSplit(node) ){
      ++numInterRepeats;
      vector<Node*> inNodes  = VUtils::getInNodes( node );
      vector<Node*> outNodes = VUtils::getOutNodes( node );
      Node* primaryInNode    = VUtils::getMaxDensityNode( inNodes );
      Node* secondaryInNode  = VUtils::getMinDensityNode( inNodes );
      Node* primaryOutNode   = VUtils::getMaxDensityNode( outNodes );
      Node* secondaryOutNode = VUtils::getMinDensityNode( outNodes );
      setNodeStatus(node, true); setUniqueness(node, false);
      if( getNodeID(primaryInNode) != getNodeID(primaryOutNode) ){ // Taking into account the knock-turn exception
	setNodeStatus(primaryInNode, true); setUniqueness(primaryInNode, true); setNodeStatus(primaryOutNode, true); setUniqueness(primaryOutNode, true);
	pushNeighboursInterRepeat( primaryInNode, node, primaryOutNode );
      }
      if( getNodeID(secondaryInNode) != getNodeID(secondaryOutNode) ){ // Taking into account the knock-turn exception
	setNodeStatus(secondaryOutNode, true); setUniqueness(secondaryOutNode, true); setNodeStatus(secondaryOutNode, true); setUniqueness(secondaryOutNode, true);
	pushNeighboursInterRepeat( secondaryInNode, node, secondaryOutNode );
      }
    }
  }
  resetNodeFlags();
  return numInterRepeats;
}

bool MetaGraph::pushNeighboursInterRepeat( Node* inNode, Node* repNode, Node* outNode ){
  NodeList* path = allocateNodeList();
  path->node = repNode;
  path->next = allocateNodeList();
  path->next->node = outNode;
  path->next->next = NULL;
  while( path ){
    Node* candidate = path->node;
    NodeList* tmpPath = path->next;
    deallocateNodeList( path );
    path = tmpPath;
    if( getUniqueness(candidate) ){
      concatenateReadStarts( inNode, candidate, graph );
      concatenateLongReads( inNode, candidate, graph );
      absorbExtensionInterRepeat( inNode, candidate );
      for( int cat = 0 ; cat<CATEGORIES ; ++cat ){
	incrementVirtualCoverage( inNode, (Category)cat, getVirtualCoverage(candidate, (Category)cat) );
	incrementOriginalVirtualCoverage( inNode, (Category)cat, getOriginalVirtualCoverage(candidate, (Category)cat) );
      }
      destroyNode( candidate, graph );
      return true;
    } else {
      adjustShortReadsInterRepeat( inNode, candidate );
      adjustLongReads( inNode, getNodeLength(candidate) );
      absorbExtensionInterRepeat( inNode, candidate );
    }
  }
  return false;
}

/*
bool MetaGraph::pushNeighboursInterRepeat( Node * node, Node * nodeInterRepeat, Node * oppositeNode ){
  Node *candidate;
  Node *lastCandidate = NULL;
  Category cat;
  NodeList *path, *tmp;
	
  // Make path (= NodeList of node and oppositeNode)
  path = allocateNodeList();
  path->node = nodeInterRepeat;
  path->next = allocateNodeList();
  path->next->node = oppositeNode;
  path->next->next = NULL;
  // Original
  while (path) {
    candidate = path->node;
    tmp = path->next;
    deallocateNodeList(path);
    path = tmp;
    
    if (getUniqueness(candidate)) {
      concatenateReadStarts(node, candidate, graph);
      concatenateLongReads(node, candidate, graph);
      absorbExtensionInterRepeat(node, candidate);
      // Read coverage
      for (cat = 0; cat < CATEGORIES; cat++) {
	incrementVirtualCoverage(node, cat, getVirtualCoverage(candidate, cat));
	incrementOriginalVirtualCoverage(node, cat, getOriginalVirtualCoverage(candidate, cat));
      }

      printf("\tConcatenated InNode %d -- OutNode %d\n", getNodeID(node), getNodeID(candidate));
      destroyNode(candidate, graph);
      return true;
    } else {
      adjustShortReadsInterRepeat(node, candidate);
      adjustLongReads( node, getNodeLength(candidate) );
      absorbExtensionInterRepeat(node, candidate);
      lastCandidate = candidate;
    }
  }
  return false;
}
*/

void MetaGraph::absorbExtensionInterRepeat( Node * node, Node * extension ){
  appendNodeGaps( node, extension, graph );
  appendDescriptors( node, extension ); // cause error ... ?
  while (getArc(node) != NULL){
    destroyArc( getArc(node), graph );
  }
  // Create new
  for( Arc* arc = getArc(extension) ; arc != NULL ; arc = getNextArc(arc) ){
    createAnalogousArc( node, getDestination(arc), arc, graph );
  }
}

void MetaGraph::adjustShortReadsInterRepeat( Node * target, Node * source ){
  if( !readStartsAreActivated(graph) ){
    return;
  }
  ShortReadMarker* targetArray = getNodeReads( getTwinNode(target), graph );
  IDnum targetLength = getNodeReadCount( getTwinNode(target), graph );
  Coordinate nodeLength = getNodeLength( source );
  
  for( IDnum index=0 ; index<targetLength ; ++index ){
    ShortReadMarker* marker = getShortReadMarkerAtIndex( targetArray, index );
    Coordinate position = getShortReadMarkerPosition( marker );
    position += nodeLength;
    setShortReadMarkerPosition( marker, position );
  }
}

void MetaGraph::forceSeparateChimericSubgraph( double primaryCoverage, double secondaryCoverage ){
  for( IDnum i=0; i<getSubgraphMaskSize() ; ++i){
    if( subgraphMask[i] == META_GRAPH_MASK_NOW_VISITING ){
      Node* node = getNodeInGraph( graph, i - nodeCount(graph) );
      if (node == NULL)
	continue;
      double curCoverage = VUtils::getNodeDensity(node);
      if( fabs(curCoverage - primaryCoverage) <= fabs(curCoverage - secondaryCoverage)){
	subgraphMask[i] = META_GRAPH_MASK_PRIMARY;
      } else {
	subgraphMask[i] = META_GRAPH_MASK_SECONDARY;
      }
    }
  }
}

void MetaGraph::identifyUniqueNodesSubgraph( double primaryCoverage ){
  for( IDnum i=0; i<nodeCount(graph) ; ++i ){
    Node* node = getNodeInGraph( graph, i+1 );
    if( node != NULL ){
      if( subgraphMask[i+1+nodeCount(graph)] == META_GRAPH_MASK_PRIMARY ){
	setUniqueness( node, isUniqueSubgraph(node, primaryCoverage) );
      } else {
	setUniqueness(node, false);
      }
    }
  }
}

bool MetaGraph::isUniqueSubgraph( Node * node, double expCoverage ){
  int nodeLength = getNodeLength(node);
  if (nodeLength > META_GRAPH_LONG_NODE_CUTOFF) {
    double nodeDensity = VUtils::getNodeDensity( node );
    double probability = -1 * META_GRAPH_LN2/2 + nodeLength/(2*expCoverage) * (expCoverage*expCoverage - nodeDensity*nodeDensity/2);
    return probability > META_GRAPH_UNIQUE_PROBABILITY_CUTOFF;
  }
  return false;
}

void MetaGraph::readCoherentSubgraph( double primaryCoverage ){
  RecycleBin* listMemory = newRecycleBin(sizeof(PassageMarkerList), 100000);
  if( !trimLongReadTipsSubgraph() ){
    destroyRecycleBin(listMemory);
    return;
  }
  IDnum previousNodeCount = 0;
  int checkModified = -1;
  while( previousNodeCount != getNumActiveNodes() ){
    previousNodeCount = getNumActiveNodes();
    for( IDnum nodeIndex=1 ; nodeIndex<=nodeCount(graph) ; nodeIndex++ ){
      Node* node = getNodeInGraph(graph, nodeIndex);
      if( node == NULL || !getUniqueness(node) ){
	continue;
      }
      while( uniqueNodesConnect(node) ){
	node = bypass();
      }
      node = getTwinNode(node);
      while( uniqueNodesConnect(node) ){
	node = bypass();
      }
    }
    checkModified++;
  }
  destroyRecycleBin(listMemory);
}

bool MetaGraph::trimLongReadTipsSubgraph(){
  bool flagLongRead = false;
  for( IDnum index=1 ; index<=nodeCount(graph) ; index++ ){
    Node* node = getNodeInGraph(graph, index);
    if (node == NULL || subgraphMask[index + nodeCount(graph)] != 1){
      continue;
    }
    if( getUniqueness(node) ){
      continue;
    }
    PassageMarkerI marker, next;
    for( marker=getMarker(node) ; marker!=NULL_IDX ; marker=next ){
      next = getNextInNode(marker);
      flagLongRead = true;
      if( !isInitial(marker) && !isTerminal(marker) ){
	continue;
      }
      if( isTerminal(marker) ){
	marker = getTwinMarker(marker);
      }
      while( !getUniqueness(getNode(marker)) ){
	if( next!=NULL_IDX && (marker==next || marker==getTwinMarker(next)) ){
	  next = getNextInNode(next);
	}
	if( getNextInSequence(marker) != NULL_IDX ){
	  marker = getNextInSequence(marker);
	  destroyPassageMarker( getPreviousInSequence(marker) );
	} else {
	  destroyPassageMarker( marker );
	  break;
	}
      }
    }
  }
  return flagLongRead;
}

void MetaGraph::changeSubgraphMask( int fromID, int toID ){
  for( long i=0 ; i<getSubgraphMaskSize() ; ++i ){
    if( subgraphMask[i] == fromID )
      subgraphMask[i] = toID;
  }
}

long MetaGraph::getSubgraphMaskSize() const {
  return 2 * nodeCount(graph) + 1;
}

bool MetaGraph::checkLongReadExistence() const {
  for( int i=1 ; i<=nodeCount(graph) ; ++i ){
	//cout<<"tes tes before getNodeInGraph"<<endl;
    Node* node = getNodeInGraph(graph, i);
	//cout<<"tes tes after getNodeInGraph"<<endl;
	if (node==NULL) continue;
    if( getMarker(node) != NULL_IDX ){
      return true;
    }
  }
  return false;
}

void MetaGraph::resetNodeFlags(){
  resetUniqueness();
  resetNodeStatus( graph );
}

void MetaGraph::resetUniqueness(){
  for( long i=1 ; i<=nodeCount(graph) ; i++ ){
    Node* node = getNodeInGraph( graph, i );
    if( node == NULL )
      continue;
    setUniqueness( node, false );
  }
}

IDnum MetaGraph::getNumActiveNodes() const {
  IDnum count = 0;
  for( IDnum i=1 ; i<=nodeCount(graph) ; i++ ){
    Node* node = getNodeInGraph( graph, i );
    if (node != NULL){
      count++;
    }
  }
  return count;
}

void MetaGraph::saveSubgraph( size_t roundID, double cov ) const {
  if( !flagReportSubgraph ) {
    return ;
  }
  FastaWriter* writer = new FastaWriter( "[SubgraphWriter ]" );
  writer->open( subgraphPrefix + "__" + Utils::itoa(roundID) + "__" + Utils::dtoa(cov) );
  for( IDnum i=1; i<=nodeCount(graph) ; ++i ){
    Node* node = getNodeInGraph( graph, i );
    if( node == NULL )
      continue;
    if( getUniqueness( node ) ){
      FastaSeq* seq = FastaSeq::instantiate( VUtils::getTitle(node), VUtils::getSequence(node) );
      writer->send( seq );
      delete seq;
    }
  }
  writer->close();
  delete writer;
}

